@rem = '--*-Perl-*--
@echo off
if "%OS%" == "Windows_NT" goto WinNT
perl -x -S "%0" %1 %2 %3 %4 %5 %6 %7 %8 %9
goto endofperl
:WinNT
perl -x -S %0 %*
if NOT "%COMSPEC%" == "%SystemRoot%\system32\cmd.exe" goto endofperl
if %errorlevel% == 9009 echo You do not have Perl in your PATH.
if errorlevel 1 goto script_failed_so_exit_with_non_zero_val 2>nul
goto endofperl
@rem ';
#!perl
#line 15
#
#  MapHacker - (c) 2007 Stu Redman
#  Tool for Duke Nukem map hacking
#  list sprites from maps and their coordinates
#  patch maphacks into maps
#  check maphacks for duplicate entries
#  diff maps and update maphacks from the diff 
#  (so you can adjust angles directly in Mapster32)
#

use strict;

my $Version = 'MapHacker 1.0';
my $mapsubdir = 'hackmaps';
my %tilenames;
# Raw blocks of a map (without the sprites)
my ($r_header, $r_sectors, $r_numwalls, $r_walls, $r_numsprites);

my $arg = lc shift or Help();
if ($arg eq 'maps') {
  $arg = lc shift or Help();
  MakeMaps($arg);
}
elsif ($arg eq 'hacks') {
  MakeHacks();
}
else {
  Help();
}

sub MakeMaps($) {
  my $arg = shift;
  my $allsprites = $arg eq 'all';
  my %spritestolist;
  unless ($allsprites) {  # make hash of sprites to list
    foreach(split( /,/, $arg)) {
      $spritestolist{$_} = 1;
    }
  }

  # this is a safety check to prevent people from destroying their work 
  # remove it at your own risk ! :-)
  die "patch map directory $mapsubdir already present, delete it first !\n" if -d $mapsubdir;
  mkdir $mapsubdir;
  open (LIST, '>spritelist.txt');
  
  my @mapnames = glob('*.map');
  ReadNames();
  foreach my $mapname (sort @mapnames) {
    # read Map
    open(IN, '<:raw', $mapname) or die "can't open $mapname: $!";
    my $numsprites = ReadMap(\*IN, $mapname);

    # read Maphacks
    my $hackname = $mapname;
    $hackname =~ s/\.map/\.mhk/i;
    my %maphacks;
    my %allmaphacks;
    open(MHK, $hackname);
    my $line = 0;
    while(<MHK>) { # sprite 22 angoff -256
      $line++;
      if (/^\s*sprite\s+(\d+)/) {
        if ($allmaphacks{$1}) {
          printf "$hackname: duplicate maphack for sprite $1 (line $allmaphacks{$1}, line $line)\n";
        }
        elsif ($1 >= $numsprites) {
          printf "$hackname: maphack for sprite $1 (line $line), but only $numsprites sprites in map\n";
        }  
        else {
          $allmaphacks{$1} = $line;
          $maphacks{$1} = $2 if /^\s*sprite\s+(\d+)\s+angoff\s+([-\d]+)/;
        }
      }
    }
    close MHK;
    #printf "$hackname: %d angle hacks\n", scalar keys %maphacks;
    
    # //44 bytes
    #typedef struct
    #{
    # long x, y, z;
    # short cstat, picnum;
    # signed char shade;
    # char pal, clipdist, filler;
    # unsigned char xrepeat, yrepeat;
    # signed char xoffset, yoffset;
    # short sectnum, statnum;
    # short ang, owner, xvel, yvel, zvel;
    # short lotag, hitag, extra;
    #} spritetype;
    my $spritetype = 'lllssccccCCccssssssssss';
    my @r_sprites;
    my $listed = 0;
    for (my $i = 0; $i < $numsprites; $i++) {
      my $r_sprite;
      read(IN, $r_sprite, 44);
      my @spr = unpack($spritetype, $r_sprite);
      my $sx = $spr[0];
      my $sy = $spr[1];
      my $spic = $spr[4];
      my $sang = $spr[15];
      if ($maphacks{$i}) {
        $sang += $maphacks{$i};
        $sang += 2048 if $sang < 0;
        $sang -= 2048 if $sang >= 2048;
        #printf "%d org %d  new %d\n", $i, $spr[15], $sang;
        $spr[15] = $sang;
        $r_sprite = pack($spritetype, @spr);
      }
      if ($allsprites && $spic > 10 || $spritestolist{$spic}) {
        print LIST "\n$mapname:\n" unless $listed++;
        printf(LIST "%4d x %6d y %6d  ang %5d  pic %4d %s\n", $i, $sx, $sy, $sang, $spic, $tilenames{$spic});
      }
      $r_sprites[$i] = $r_sprite;
    }
    die "EOF not reached for $mapname" if read(IN, my $dummy, 1); # this should not happen
    close IN;
    if ($listed) { # we found sprites from the selection in the map, so write it
      my $newmapname = "$mapsubdir/$mapname";
      open(OUT, '>:raw', $newmapname) or die "can't open $newmapname: $!";
      print OUT $r_header, $r_sectors, $r_numwalls, $r_walls, $r_numsprites;
      for (my $i = 0; $i < $numsprites; $i++) {
        print OUT $r_sprites[$i];
      }
      close OUT;
    }
  }
  close LIST;
}

sub MakeHacks() {
  my @mapnames = glob("$mapsubdir/*.map") or die "no maps found in $mapsubdir\n";
  ReadNames();
  foreach my $mapname (sort @mapnames) {
    my $original_mapname = $mapname;
    $original_mapname =~ s-.+/--;
    my $hackname = $original_mapname;
    $hackname =~ s/\.map/\.mhk/i;

    # open original map
    open(INorg, '<:raw', $original_mapname) or die "can't open $original_mapname: $!";
    my $numsprites = ReadMap(\*INorg, $original_mapname);
    
    my @SpriteAngleOffsets;
    my @SpritePics;
    my $spritetype = 'lllssccccCCccssssssssss';
    my $hacks = 0;
    my $lastsprite;
    # calculate angle offsets for all sprites
    for (my $i = 0; $i < $numsprites; $i++) {
      my $r_sprite;
      read(INorg, $r_sprite, 44);
      my @spr = unpack($spritetype, $r_sprite);
      $SpritePics[$i] = $spr[4];
      $SpriteAngleOffsets[$i] = $spr[15];
      # Mapster changes the sprite order when saving maps (reverse order),
      # so determine if it's reversed or not.
      $lastsprite  = "$spr[4] $spr[0] $spr[1]" if ($i == $numsprites - 1);
    }
    close INOrg;

    # open patched map
    open(INnew, '<:raw', $mapname) or die "can't open $mapname: $!";
    if ($numsprites != ReadMap(\*INnew, $mapname)) {
      warn "mismatch in sprite numbers between $original_mapname and $mapname\n";
      next;
    }
    
    my $reversed = 0;
    for (my $i = 0; $i < $numsprites; $i++) {
      my $r_sprite;
      read(INnew, $r_sprite, 44);
      my @spr = unpack($spritetype, $r_sprite);
      $reversed = "$spr[4] $spr[0] $spr[1]" eq $lastsprite if ($i == 0);
      my $index = $reversed ? $numsprites - $i - 1 : $i;
      if ($spr[4] == $SpritePics[$index]) { # check if something went wrong (sprites got swapped)
        #printf "%d org %d  new %d  diff %d\n", $index, $SpriteAngleOffsets[$index], $spr[15], $spr[15] - $SpriteAngleOffsets[$index];
        $SpriteAngleOffsets[$index] = $spr[15] - $SpriteAngleOffsets[$index];
        $hacks++ if $SpriteAngleOffsets[$index];
      }
      else {
        print "$mapname: changed sprite $i\n";
      }
    }
    #print "$hackname:  $hacks\n";
    close INNew;
    
    # open maphack file (no problem if it isn't there) to preserve structure and comments
    open(MHK, $hackname);
    my @oldhacks = <MHK>;
    close MHK;
    # rebuild it
    my @mhk;
    my @hacksprocessed;
    # write preexisting hacks with adjusted angles
    my $changedhacks;
    foreach (@oldhacks) {
      if (/^\s*sprite\s+(\d+)\s+angoff\s+([-\d]+)(.*)/) {
        unless ($hacksprocessed[$1]     # drop double hack
                || $1 >= $numsprites) { # drop hacks out of range
          push @mhk, "sprite $1 angoff $SpriteAngleOffsets[$1]$3\n";
          $hacksprocessed[$1] = 1;
          my $anglediff = $2 - $SpriteAngleOffsets[$1];
          $changedhacks++ if $anglediff % 2048 != 0;
          #print "changed sprite $1 num $numsprites\n" if $anglediff % 2048 != 0;
        }
      }
      else {
        push @mhk, $_;
      }
    }
    # write hacks that were not in the mhk before
    my $newhacks = 0;
    my %nhk;
    for (my $i = 0; $i < $numsprites; $i++) {
      if (!$hacksprocessed[$i] && $SpriteAngleOffsets[$i]) {
        $nhk{sprintf("%5d %d", $SpritePics[$i], $i)}
           = "sprite $i angoff $SpriteAngleOffsets[$i] // $SpritePics[$i] $tilenames{$SpritePics[$i]}\n";
        $newhacks++;
      }
    }
    
    if ($changedhacks || $newhacks) {
      printf "$hackname:  %d new hacks, %d changed hacks\n", $newhacks, $changedhacks;
      # something changed, so open it now for writing
#      open(MHK, ">$mapsubdir/$hackname") or die "can't open $hackname for writing: $!"; # for testing
      open(MHK, ">$hackname") or die "can't open $hackname for writing: $!";
      print MHK @mhk;
      if ($newhacks) {
        print MHK "\n";
        foreach (sort keys %nhk) { # sort hacks by tile
          print MHK $nhk{$_};
        }
      }
      close MHK;
    }
  }
}

sub Help() {
  print "$Version (c) 2007 Stu Redman\n";
  print "maphacker maps all                  - make patchmaps and list all sprites\n";
  print "maphacker maps 603,1294,1323,1334   - make patchmaps and list some sprites\n";
  print "maphacker hacks                     - diff maps and update maphacks\n";
  exit(0);
}

sub ReadNames() {
  open(NAMESH, 'names.h');
  while(<NAMESH>) { ##define FEM9 3450
    $tilenames{$2} = $1 if /\#define\s+(\w+)\s+(\d+)/;
  }
  close NAMESH;
}

sub ReadMap($) {
  my $IN = shift;
  my $mapname = shift; 
  read($IN, $r_header, 22);
  my @header = unpack('llllsss', $r_header);
  die "$mapname is no Build map" if $header[0] != 7;
  #print "$header[6] sectors\n";
  read($IN, $r_sectors, $header[6] * 40);
  my $numwalls;
  read($IN, $r_numwalls, 2);
  ($numwalls) = unpack('s', $r_numwalls);
  #print "$numwalls walls\n";
  read($IN, $r_walls, $numwalls * 32);
  my $numsprites;
  read($IN, $r_numsprites, 2);
  ($numsprites) = unpack('s', $r_numsprites);
  #print "$numsprites sprites\n";
  return $numsprites;
} 
 
__END__
:endofperl
